import re
import logging
from html import unescape
from odoo.osv import expression
from odoo.upgrade import util
from odoo.tools.safe_eval import safe_eval

_logger = logging.getLogger(__name__)


def adapt_context(context_str, old, new):
    eval_context = util.SelfPrintEvalContext()
    def_old = "default_{}".format(old)
    def_new = "default_{}".format(new)

    def adapt_value(key, value):
        if key == "orderedBy" and isinstance(value, dict):
            # only adapt the "name" key
            return {k: (adapt_value(None, v) if k == "name" else v) for k, v in value.items()}

        if not isinstance(value, str):
            # ignore if not a string
            return value

        parts = value.split(":", 1)
        if parts[0] != old:
            # if not match old, leave it
            return value
        # change to new, and return it
        parts[0] = new
        return ":".join(parts)

    def adapt_dict(d):
        # adapt (in place) dictionary values
        if not isinstance(d, dict):
            return

        for key in util.fields._CONTEXT_KEYS_TO_CLEAN:
            if d.get(key):
                d[key] = [adapt_value(key, e) for e in d[key]]

        for vt in ["pivot", "graph", "cohort"]:
            key = "{}_measure".format(vt)
            if key in d:
                d[key] = adapt_value(key, d[key])

            if vt in d:
                adapt_dict(d[vt])

    try:
        context = safe_eval(context_str, eval_context, nocopy=True)
        adapt_dict(context)
        if def_old in context:
            context[def_new] = context.pop(def_old)
        context_new = str(context)
        return context_new
    except Exception as e:
        _logger.warning("Unable to adapt context (old '%s' -> new '%s') got '%s' for %s",
            old, new, str(e), context_str
        )
        return None

def get_field_relation(cr, model, fname):
    cr.execute(
        """
        SELECT relation, id
        FROM ir_model_fields
        WHERE model = %s
        AND name = %s
        ORDER BY id
        LIMIT 1
        """,
        [model, fname],
    )
    return cr.fetchone() if cr.rowcount else (None, None)


def migrate(cr, version):
    # https://github.com/odoo/enterprise/pull/46621:
    # - the sale.order `recurrence_id` field is removed and a new field
    # `plan_id` is introduced and now hold the "recurrence" (duration, unit)
    # information.
    #   As we have a lot of dashboards, filters or export from users that use
    # the former field, replace it by `plan_id` when possible.
    # Note: we only cover ir.ui.view.custom, ir.filters, ir.export
    target_model = "sale.order"
    old = "recurrence_id"
    new = "plan_id"

    match_old = r"\y{0}\y".format(old)

    def _adapt_recurrence_domain(leaf, is_or, negated):
        recc_name_val = {
            3: "Monthly",
            5: "1 Year",
            6: "3 Years",
            7: "5 Years",
            8: "2 Years",
            9: "4 Years",
        }

        left, op, right = leaf
        if op in ("=", "!=") and isinstance(right, int) and right in recc_name_val:
            # likewise: ('recurrence_id', '=', 3)
            op, right = "ilike" if op == "=" else "not ilike", recc_name_val[right]
        if (
            op in ("in", "not in") and isinstance(right, (list, tuple))
            and len(right) == 1 and right[0] in recc_name_val
        ):
            # likewise: ('recurrence_id', 'in', [3])
            op = "ilike" if op == "in" else "not in"
            right = recc_name_val[right[0]]

        if op in ("=", "!=") and isinstance(right, str):
            # fixup wrong leaf
            op = "ilike" if op == "=" else "not ilike"

        if op in ("ilike", "not ilike"):
            res = [(left.replace("recurrence_id", "plan_id"), op, right)]
            return res
        _logger.debug("Unable to adapt 'recurrence_id' in domain leaf: %s" % (leaf,))

        # replace by TRUE_LEAF unless negated
        if is_or ^ negated:
            return [expression.FALSE_LEAF]
        return [expression.TRUE_LEAF]

    #
    # 1. Adapt dashboards
    #
    for id_, act in util.helpers._dashboard_actions(cr, r"\yrecurrence_id\y"):
        try:
            act_id = int(act.get("name", "FAIL"))
        except ValueError:
            continue

        cr.execute("SELECT res_model FROM ir_act_window WHERE id = %s", [act_id])
        if not cr.rowcount:
            continue
        [act_model] = cr.fetchone()
        changed = False

        if act_model == target_model:
            context = adapt_context(act.get("context") or "{}", old, new)
            if context:
                act.set("context", context)
                changed = True

        domain = act.get("domain")
        if any(entity in domain for entity in ("&#27;", "&amp;", "&lt;", "&gt;")):
            # There is a bug Odoo 16.0 that double escape the domains in dashboad...
            # See https://github.com/odoo/odoo/pull/119518
            domain = unescape(domain)
        domain = util.domains._adapt_one_domain(
            cr, target_model, old, new, act_model, domain,
            adapter=_adapt_recurrence_domain,
        )
        if domain:
            act.set("domain", str(domain))
            changed = True

        if changed:
            util.add_to_migration_reports(
                ("ir.ui.view.custom", id_, act.get("string", "ir.ui.view.custom")),
                "Filters/Dashboards"
            )

    #
    # 2. Adapt ir.filters
    #
    cr.execute("""
        SELECT id, name, model_id, domain, context, sort
        FROM ir_filters
        WHERE domain ~ %s OR context ~ %s OR sort ~ %s
    """, (match_old, match_old, match_old))
    for f_id, f_name, f_model, f_domain, f_context, f_sort in cr.fetchall():
        updated_fields, params = [], []
        if old in f_sort:
            updated_fields.append('sort')
            params.append(f_sort.replace(old, new))
        if old in f_context:
            context = adapt_context(f_context or "{}", old, new)
            if context:
                updated_fields.append("context")
                params.append(context)
        if f_model == target_model and old in f_domain:
            domain = util.domains._adapt_one_domain(
                cr, target_model, old, new, f_model, f_domain,
                adapter=_adapt_recurrence_domain,
            )
            if domain:
                updated_fields.append("domain")
                params.append(str(domain))

        if updated_fields:
            query = "UPDATE ir_filters SET {0} WHERE id = %s".format(
                ", ".join("{0} = %s".format(f) for f in updated_fields)
            )
            cr.execute(query, tuple(params + [f_id]))
            util.add_to_migration_reports(
                ("ir.filters", f_id, f_name),
                "Filters/Dashboards"
            )

    #
    # 3. Adapt ir.exports.line
    #
    cr.execute("""
        SELECT l.id, e.resource, l.name
        FROM ir_exports_line l
        JOIN ir_exports e ON (l.export_id = e.id)
        WHERE l.name ~ %s
    """, (match_old,))
    for export_line_id, export_model, export_fields_path, in cr.fetchall():
        current_model = export_model
        paths = []
        for field in export_fields_path.split("/"):
            if not current_model:
                paths.append(field)
                continue
            field_relation, field_id = get_field_relation(cr, current_model, field)
            if not field_id:
                paths.append(field)
                continue
            if current_model == target_model and field == old:
                field = new
            paths.append(field)
            current_model = field_relation
        new_export_fields_path = "/".join(paths)
        if new_export_fields_path != export_fields_path:
            cr.execute(
                "UPDATE ir_exports_line SET name = %s WHERE id = %s",
                (new_export_fields_path, export_line_id)
            )
